// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
/* written by John Doug Reynolds, February 2002 */
/* Updates, April 2005, Jenny Carter */

/*
 * AI: 26/1/2007 
 * General remark: All the float calculations have been converted into double.
 * Also, rgslib ahs been modified in orden to user double in some of the calculations.
 */

#include "rgsfluxer.h"
#include "error.h"
#include "param.h"
#include <rgsPixelsLib.h>
#include <DalVec.h>
#include <cstdio>
#include <cmath>
#include "PhysicalConstants.h"

using namespace rgsPixelsLib;
using namespace std;

const double hc = PhysicalConstants::eVtimesAo/1000.; //12.39854; // [keV Angstrom] energy times wavelength UPDATED 07/05/2005

class BetaToLambda {
public:

	BetaToLambda( double alphaNom, double lineSep, unsigned order );
	double operator()( double beta ) const { return _slope * cos(beta) + _icept; }
	double getOrder();
private:
	double _slope, _icept,_order;
};

class ResponseMatrix {
public:
	~ResponseMatrix();
	ResponseMatrix( const Table* );
	pair<double,double> changeRow( unsigned );
	bool next( short &channel, double &value );
	unsigned rows() const { return _rows; }
private:
	const DalVecI _nGrp;
	const DalVecE _eLow, _eHigh, _arfData;
	const Column *_fChan, *_nChan, *_matrix;
	const CellData *_fChanData, *_nChanData, *_matrixData;
	const short *_fChanPtr, *_nChanPtr;
	const float *_matrixPtr;
	unsigned _groupCount, _itemCount,_rows;
	short _channel;
};

class Fluxer {
public:
	enum AxisMode { Wavelength, Energy };

	Fluxer( enum AxisMode, const GridSpan&, const vector<int> &qualities, bool witharf=false );

	void add( const string &pha,  const string &bkg, const string &rmf,const string &arf, bool &withbkg );

	bool at( int cell, float &chan, float &flux, float &err, int &qual ) const;

	map<int,double> specMapGridMin, specMapGridMax;

private:
	enum AxisMode _mode;
	struct GridSpan _grid;
	bool _witharf;
	vector<float> _rate, _err2, _area, _resp, errorvalues;
	vector<int> _quality;
	unsigned _qualityRank[5];


	static void rebin( vector<float> &vec, float min, float max, float val );

	double binWidth( const pair<double,double>& ) const;
	double energyToGrid( double energy ) const;
	double lambdaToGrid( double lambda ) const;
	unsigned addSpectrum( const string &filename, const string &bfilename, vector<bool> &bad, bool &withbkg );
	void addResponse( const string &filename,const string &arffilename,
			unsigned spectrumOrder, const vector<bool> &bad );
	void includeBack(const string &filename, vector<double> &cvalues, vector<double> &bvalues, vector<double> &errorvalues, double exposure );
};

void rgsfluxer()
{   
	string filename = stringParameter("file");
	string format = stringParameter("format");
	bool withaltnan = booleanParameter("withaltnan");
	bool witharf = booleanParameter("witharf");
	double nan = realParameter("nan");
	bool withbkg = false;

	vector<string> pha = stringListParameter("pha");
	vector<string> rmf = stringListParameter("rmf");
	vector<string> arf = stringListParameter("arf");
	vector<int> qualities = intListParameter("quality");

	vector<string> bkg = stringListParameter("bkg"); // New parameter 


	//Check if the user has changed the default lambda grid values
	//4-40 Ang --> 3600 channels...
	//If this is the case, we will automatically change those values
	//to 3.99
	struct GridSpan grid( realParameter("min"),
			realParameter("max"),
			intParameter("bins")
	);

	//Check if all the input spectra have the same energy grid.
	//If so, we will create exactly the same grid for the output flux.
	//This feature is controlled by the default grid parameter.
	if (booleanParameter("tryspectrumgrid"))
	{
		errstr << "Checking if all the spectra grids are aligned"
				<< Sas::message(Msg::App,Msg::Verbose);
		DataSet *set=0;
		struct GridSpan spectrumGrid;
		std::vector<GridSpan> checkGrids;
		bool spectraaligned = true;
		std::vector<string> files2check = pha;
		files2check.insert(files2check.end(), bkg.begin(), bkg.end());

		for (unsigned int i = 0 ; i < files2check.size() ; ++i)
		{
			set = dataSetServer->open(files2check[i],Dal::Read);
			Table *table = set->table("SPECTRUM");
			DalVecI channel( table->column("CHANNEL") );
			if (table->column("CHANNEL")->attribute("TCUNI")->asString() != "Angstrom")
			{
				errstr << "SPECTRUM with beta grid. Skip possible lambda grid alignment"<<Sas::warning("BetaGrid");
				spectraaligned = false;
				break;
			}
			struct GridSpan spectrumGrid = gridspan( channel.column() );
			spectrumGrid = GridSpan( Grid( spectrumGrid.stride, spectrumGrid[1] ),
					Span(spectrumGrid.min-1,spectrumGrid.max-1) );
			checkGrids.push_back( spectrumGrid );
			if (checkGrids.size() == 2)
			{
				if (!commensurate(checkGrids[0],checkGrids[1]))
				{
					errstr << "Different input spectral grid and default grid. Rebinning method will be applied "
							<<Sas::warning("smoothing");
					spectraaligned = false;
				}
				checkGrids.pop_back();
			}
		}

		if (spectraaligned == true)
		{
			errstr << "All spectra grid are aligned. Input spectrum grid will be used"
					<< Sas::message(Msg::App,Msg::Verbose);
			grid = checkGrids[0];
		}

		dataSetServer->close(set);
	}
	else
	{
		errstr << "User's defined spectrum grid will be used"
			<< Sas::message(Msg::App,Msg::Verbose);
	}


	enum Fluxer::AxisMode mode=(Fluxer::AxisMode)-1;
	string modeStr = stringParameter("mode");
	const char *binUnits=0, *valUnits=0;
	if ( modeStr == "wavelength" ) {
		mode = Fluxer::Wavelength;
		binUnits = "Angstrom";
		valUnits = "1/(s cm^2 A)";
	}
	else if ( modeStr == "energy" ) {
		mode = Fluxer::Energy;
		binUnits = "keV";
		valUnits = "1/(s cm^2 keV)";
	}
	else Sas::error("badMode","illegal value specified for mode parameter");

	if ( pha.size() != rmf.size() )
		Sas::error("pairing","unpaired pha or rmf");

	struct Fluxer fluxer( mode, grid, qualities,witharf );

	vector<string>::const_iterator phaI = pha.begin(), rmfI = rmf.begin(), arfI=arf.begin();

	vector<string>::const_iterator bkgI = pha.begin();

	if (bkg.size() || 0){	
		withbkg = true;
	}
	if (withbkg){
		if ( pha.size() != bkg.size() ){
			Sas::error("pairing","unpaired pha or bkg");
		}
		bkgI = bkg.begin();
	}

	for ( string defaultRmf; rmfI != rmf.end(); ++rmfI,++arfI ) {
		string rmfName;
		string arfName="@";
		if ( *rmfI == "@" ) {
			rmfName = *phaI;
			string::size_type size = rmfName.size();
			if ( size < 14 || rmfName.substr(size-14,6) != "SRSPEC" )
				Sas::error("badPhaName","cannot infer corresponding rmf");
			rmfName.replace( size-14, 6, "RSPMAT" );
		}
		else if ( *rmfI == "." ) {
			if ( defaultRmf == "" )
				Sas::error("illegalRmfName","first rmf cannot be '.'");
			rmfName = defaultRmf;
		}
		else
		{
			rmfName = *rmfI;
			if (witharf)
				arfName = *arfI;
		}

		fluxer.add( *phaI++, *bkgI++, rmfName, arfName, withbkg );
		defaultRmf = rmfName;
	}

	//

	bool withquality = (qualities.size() > 1);

	if ( format == "dal" ) {
		if ( filename == "" ) filename = "flux.ds";
		struct DataSet *dataset = dataSetServer->open( filename, Dal::Create );
		struct Table *table = dataset->addTable("FLUXED",grid.length());	 
		if ( withbkg)
			*table->addAttribute("BACKIN") = true; // Attribute to header to show bkg calculation used
		{
			DalVecE channel( table->addColumn("CHANNEL",Dal::Real32,
					modeStr.c_str(),binUnits) );
			DalVecE flux( table->addColumn("FLUX",Dal::Real32,
					"observed flux",valUnits) );
			DalVecE error( table->addColumn("ERROR",Dal::Real32,
					"statistical uncertainty",valUnits) );  
			DalVecI quality;
			if ( withquality )
				quality.reset( table->addColumn("QUALITY",Dal::Int16,
						"quality flag") );

			for ( int c = grid.min; c <= grid.max; ++c ) {
				int qual, i = c - grid.min;
				if ( !fluxer.at(c,channel[i],flux[i],error[i],qual) )
				{
					if ( !withaltnan ) {
						flux.setToNull(i);
						error.setToNull(i);
					}
					else flux[i] = error[i] = nan;
				}
				if (flux[i] == 0.0){
					flux.setToNull(i);
					error.setToNull(i);
				}
				if ( withquality ) quality[i] = qual;
			}
		}
		dataSetServer->close( dataset );
	}
	else {
		if ( filename == "" )
			filename = format == "qdp" ? "flux.qdp" : "flux.txt";
		const char *snan = format == "qdp" ? "no" : "NaN";
		FILE *out = fopen( filename.c_str(), "w" );
		if ( format == "qdp" ) fprintf( out, "READ SERR 2\n" );
		fprintf( out, "!%14s%15s%15s%10s\n", modeStr.c_str(),
				"flux", "error", withquality ? "quality" : "" );
		fprintf( out, "!%14s%15s%15s\n", binUnits, valUnits, valUnits );
		for ( int c = grid.min; c <= grid.max; ++c ) {
			float chan, flux = nan, err = nan;
			int qual;
			if ( fluxer.at(c,chan,flux,err,qual) || withaltnan )
				fprintf( out, "%15e%15e%15e", chan, flux, err );
			else
				fprintf( out, "%15e%15s%15s", chan, snan, snan );
			if ( withquality )
				fprintf( out, "%10d", qual );
			fprintf( out, "\n" );
		}
		fclose( out );
	}
}

/* --- BetaToLambda methods ------------------------------------------------ */

BetaToLambda::BetaToLambda( double alphaNom, double lineSep, unsigned order )
{
	_slope = -lineSep / order;
	_icept = -cos(alphaNom) * _slope;
	_order = double(order);
}


double BetaToLambda::getOrder()
{
	return _order;
}
/* --- ResponseMatrix methods ---------------------------------------------- */

ResponseMatrix::~ResponseMatrix()
{
	delete _matrixData;
	delete _nChanData;
	delete _fChanData;
}

ResponseMatrix::ResponseMatrix( const Table *table )
: _nGrp( table->column("N_GRP") ),
_eLow( table->column("ENERG_LO") ),
_eHigh( table->column("ENERG_HI") ),
_fChan( table->column("F_CHAN") ),
_nChan( table->column("N_CHAN") ),
_matrix( table->column("MATRIX") ),
_fChanData(0), _nChanData(0), _matrixData(0),
_groupCount(0), _itemCount(0),
_rows( table->rows() )
{}

pair<double,double> ResponseMatrix::changeRow( unsigned row )
{
	delete _matrixData;
	delete _nChanData;
	delete _fChanData;

	_fChanData = _fChan->cellData(row);
	_nChanData = _nChan->cellData(row);
	_matrixData = _matrix->cellData(row);

	_fChanPtr = _fChanData->int16Data();
	_nChanPtr = _nChanData->int16Data();
	_matrixPtr = _matrixData->real32Data();

	_groupCount = _nGrp[row];
	_itemCount = 0;

	return pair<double,double>(_eLow[row],_eHigh[row]);

}

bool ResponseMatrix::next( short &channel, double &value )
{
	for (;;)
	{
		if ( _itemCount ) {
			channel = _channel++;
			value = *_matrixPtr++ ;
			-- _itemCount;
			return true;
		}
		else if ( _groupCount ) {
			_channel = *_fChanPtr++ - 1;
			_itemCount = *_nChanPtr++;
			-- _groupCount;
		}
		else return false;
	}
}

/* --- Fluxer methods ------------------------------------------------------ */

void Fluxer::rebin( vector<float> &vec, float min, float max, float val )
{ 
	if ( max < min ) swap(min,max);

	if ( min >= vec.size() || max <= 0 ) return;

	if ( min < 0 ) {
		val *= 1 + min/(max - min);
		min = 0;
	}

	if ( max > vec.size() ) {
		val *= 1 - (max - vec.size())/(max - min);
		max = vec.size();
	}

	if ( floor(max) > floor(min) ) {
		double portion = val / (max - min);

		if ( ceil(min) > min ) 
		{
			vec[int(floor(min))] += portion * (ceil(min) - min);
			min = ceil(min);
		}
		while ( floor(max) > floor(min) ) 
		{
			vec[int(floor(min))] += portion;
			min += 1;
		}
		vec[int(floor(min))] += portion * (max - min);
	}
	else
		vec[int(floor(min))] += val;
}

Fluxer::Fluxer( enum AxisMode mode, const GridSpan &grid,
		const vector<int> &qualities,  bool witharf)
: _mode(mode), _grid(grid), _witharf(witharf),
_rate(grid.length()), _err2(grid.length()),
_area(grid.length()), _resp(grid.length()),_quality(grid.length())
{
			// for (unsigned j = 0; j < grid.length(); ++j){
			// errorvalues[j] = 0.0;
			//_rate.setToNull(j); 
			//_rate[j] = NULL;
			//} 
			unsigned i;
			vector<int>::const_iterator iter;
			for ( i = 0; i < 5; ++i ) _qualityRank[i] = 0;
			for ( i = 0, iter = qualities.begin(); iter != qualities.end(); ++iter )
				_qualityRank[*iter] = ++i;
}

void Fluxer::add( const string &pha, const string &bkg, const string &rmf, const string &arf, bool &withbkg )
{  
	vector<bool> bad;
	addResponse( rmf, arf, addSpectrum(pha, bkg, bad, withbkg), bad );
}


bool Fluxer::at( int c, float &chan, float &flux, float &err, int &qual ) const
{
	int i = c - _grid.min;
	double area = _area[i];

	if ( area ) {	 
		flux = _rate[i] / area;   
		err = sqrt( _err2[i] ) / area;
	}
	qual = _quality[i];
	chan = _grid[c];
	return bool(area);
}

double Fluxer::binWidth( const pair<double,double> &limits ) const
{
	return _mode == Energy
	? limits.second - limits.first : hc/limits.first - hc/limits.second;
}

double Fluxer::energyToGrid( double energy ) const
{
	return _grid.tick( _mode == Energy ? energy : hc / energy );
}

double Fluxer::lambdaToGrid( double lambda ) const
{
	return _grid.tick( _mode == Wavelength ? lambda : hc / lambda );
}

inline double sqr( double x ) { return x * x; }
unsigned Fluxer::addSpectrum( const string &filename, const string &bfilename, vector<bool> &bad, bool &withbkg )
{
	struct DataSet *dataset = dataSetServer->open(filename,Dal::Read);
	const Table *table = dataset->table("SPECTRUM");
	unsigned order = table->attribute("RFLORDER")->asInt();
	double exposure = table->attribute("EXPOSURE")->asReal();
	bool gross = table->attribute("HDUCLAS2")->asString() != "NET";

	bad.resize( table->rows(), false );

	struct BetaToLambda toLambda( table->attribute("RGSALPHA")->asReal(),
			table->attribute("LINE_SEP")->asReal(),
			order );

	{
		const DalVecI channel( table->column("CHANNEL") );
		const DalVecJ counts( table->column("COUNTS") );
		const DalVecE areascal( table->column("AREASCAL") );    
		const DalVecI quality( table->column("QUALITY") );
		DalVecE stat_err;
		DalVecE backscal; 
		if ( !gross ){
			stat_err.reset( table->column("STAT_ERR") );
		} else {  
			backscal.reset(table->column("BACKSCAL") );
		}

		struct GridSpan beta = gridspan( channel.column() );

		vector<double> cvalues;
		vector<double> bvalues;
		vector<double> stat_errall;

		for ( unsigned i = 0; i < table->rows(); ++i ){
			cvalues.push_back(counts[i]);
			if ( gross){
				bvalues.push_back(backscal[i]);
				double value = counts[i];
				stat_errall.push_back(sqrt(value)); // Total spectra
			} else {
				stat_errall.push_back((stat_err[i])); // Net spectra
			}
		}

		if (withbkg){  
			if (!gross) {
				Sas::warning("Background subtraction from net spectrum??", 
						"You include background file(s) but spectrum already subtracted! Ignoring file(s)"); 
				withbkg = false;	
			}
			else {
				includeBack(bfilename, cvalues, bvalues, stat_errall, exposure);
			}
		}

		for ( unsigned i = 0; i < table->rows(); ++i )
			if ( _qualityRank[ quality[i] ] ) {
				float min,max;
				//AI: Wavelength scale.
				if (table->column("CHANNEL")->attribute("TCUNI")->asString() == "Angstrom")
				{
					min = lambdaToGrid((beta[channel[i]-.5])/toLambda.getOrder()) + .5;
					max = lambdaToGrid((beta[channel[i]+.5])/toLambda.getOrder()) + .5;
				}
				else
				{
					min = lambdaToGrid( toLambda(beta[channel[i]-.5]) ) + .5;
					max = lambdaToGrid( toLambda(beta[channel[i]+.5]) ) + .5;
				}

				float time = exposure * areascal[i];
				float  rate = cvalues[i] / time;

				/*
				 * Map where we stored the energy grid to be used afterward during 
				 * the calculation of the response matrix. 
				 * With this operation we are sure that we are using the same
				 * energy grid for the count rate calculation and area calculation.
				 */

				specMapGridMin[i+1]=min;
				specMapGridMax[i+1]=max;

				rebin( _rate, min, max, rate );

				//float err2 = gross ? rate / time : sqr(stat_err[i] / time); // Old calculation
				float err2 = sqr(stat_errall[i])/sqr(time); // To put the new error (error on counts) into previous structure (error on flux)

				rebin( _err2, min, max, err2 );

				struct Span span = Span(int(floor(min)),int(floor(max))) & _grid;
				for ( int c = span.min; c <= span.max; ++c ) {
					int j = c - _grid.min;	     
					if ( _qualityRank[ _quality[j] ] < _qualityRank[ quality[i] ] )
						_quality[j] = quality[i];	    
				}
			}
			else bad[i] = true;   		
	}
	dataSetServer->close( dataset );
	return order;
}
// ------
void Fluxer::includeBack(const string &filename, vector<double> &cvalues, vector<double> &bvalues, vector<double> &errorvalues, double exposure )
{   
	struct DataSet *dataset = dataSetServer->open(filename,Dal::Read);
	const Table *table = dataset->table("SPECTRUM");
	string units  = table->attribute("HDUCLAS3")->asString();
	if (units == "COUNT")
	{
		float bkgexposure = table->attribute("EXPOSURE")->asReal();
		{  
			const DalVecJ counts( table->column("COUNTS") );
			const DalVecE areascal( table->column("AREASCAL") );
			const DalVecE backscal( table->column("BACKSCAL") ); 
			for ( unsigned i = 1; i < table->rows(); ++i )
			{ 
				//	    cvalues[i] = cvalues[i] - (counts[i] * (bvalues[i]/backscal[i])); 
				// New error calculation 
				//	    errorvalues[i] = sqrt(sqr(errorvalues[i]) + ((sqr(bvalues[i]/backscal[i]))*counts[i]));

				cvalues[i] = cvalues[i] - (counts[i]*exposure/bkgexposure * (bvalues[i]/backscal[i]));
				// New error calculation 	    
				errorvalues[i] = sqrt(sqr(errorvalues[i]) + ((sqr(bvalues[i]/backscal[i]))*counts[i]*sqr(exposure/bkgexposure)));

			}
		}
	}
	else
	{
		// Then HDUCLAS3 is RATE
		const DalVecE rate( table->column("RATE") );
		const DalVecE errv( table->column("STAT_ERR") ); // error on the background
		const DalVecE areascal( table->column("AREASCAL") );
		const DalVecE backscal( table->column("BACKSCAL") );
		//	float exposure = table->attribute("EXPOSURE")->asReal();
		for ( unsigned i = 0; i < table->rows(); ++i )
		{ 
			cvalues[i] = cvalues[i] - ((rate[i] * exposure) * (bvalues[i]/backscal[i]));
			// New error calculation	 
			errorvalues[i] = sqrt(sqr(errorvalues[i]) + ((sqr(errv[i]))*(sqr(bvalues[i]/backscal[i]))*(sqr(exposure))));
		}
	}
	dataSetServer->close( dataset );
}


// ------
void Fluxer::addResponse( const string &filename, const string &arffilename,
		unsigned spectrumOrder, const vector<bool> &bad )
{
	struct DataSet *dataset = dataSetServer->open(filename,Dal::Read);
	const Table *btab = dataset->table("EBOUNDS");


	struct DataSet *arfset;
	const Table *arftab;
	DalVecE resp;
	//Default value in case we are dealing with a RSPMAT file.
	double arf = 1.0;
	if (arffilename != "@")
	{
		arfset = dataSetServer->open(arffilename,Dal::Read);
		arftab = arfset->table("SPECRESP");
		resp.reset(arftab->column("SPECRESP") );
	}




	if ( int(spectrumOrder) != dataset->attribute("RFLORDER")->asInt() )
		Sas::error("wrongOrder",
				"response matrix order not same as spectrum");

	{
		const DalVecI channel( btab->column("CHANNEL") );
		for ( unsigned i = 1; i <= btab->rows(); ++i )
			if ( channel(i) != short(i) )
				Sas::error("channelGaps",
						"cannot handle response matrix with channel gaps");
	}

	{
		struct ResponseMatrix matrix( dataset->table("MATRIX") );
		const DalVecE eMin( btab->column("E_MIN") );
		const DalVecE eMax( btab->column("E_MAX") );

		for ( unsigned i = 0; i < matrix.rows(); ++i ) 
		{
			short channel;
			double value;
			double delta = binWidth( matrix.changeRow(i));	  
			if (_witharf)
				arf = resp[i];
			while ( matrix.next(channel,value) )
			{
				if ( ! bad[channel] )
				{

					/*
					 * Now uses the energy grid calculation previously to calculate the 
					 * fraction of the are that has to be rebinned.	      
					 */
					rebin( _area,
							specMapGridMin[channel + 1],
							specMapGridMax[channel + 1],
							delta*value*arf
					);
				}
			}

		}
	}

	dataSetServer->close( dataset );
}
